// boost includes: for marshalling and unmarshalling of game commands
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>

// Lua and Sol
#include <sol.hpp>
#pragma comment(lib, "liblua53.a")

// bell0bytes util
#include "inputHandler.h"
#include "serviceLocator.h"
#include "stringConverter.h"

// bell0bytes graphics: for the custom mouse cursor
#include "sprites.h"

// bell0bytes core
#include "app.h"

namespace input
{
	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////// Constructors and Destructors //////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////

	// BindInfo
	BindInfo::BindInfo() : keyCode(0), keyState(KeyState::JustReleased) {};
	BindInfo::BindInfo(const unsigned int keyCode, const KeyState keyState) : keyCode(keyCode), keyState(keyState) {};

	// GameCommand
	GameCommand::GameCommand() : name(L""), chord(0) {};
	GameCommand::GameCommand(const std::wstring& name) : name(name), chord(0) {};
	GameCommand::GameCommand(const std::wstring& name, const unsigned int keyCode, const KeyState keyState) : name(name)
	{
		chord.push_back(BindInfo(keyCode, keyState));
	}
	GameCommand::GameCommand(const std::wstring& name, const BindInfo& bi) : name(name)
	{
		chord.push_back(bi);
	};
	GameCommand::GameCommand(const std::wstring& name, const std::vector<BindInfo>& chord) : name(name), chord(chord) {};
	
	// Keyboard and Mouse
	KeyboardAndMouse::KeyboardAndMouse() : mouseX(0), mouseY(0)
	{
		this->currentState.fill(false);
		this->previousState.fill(false);
		this->mouseCursor = nullptr;
	}
	KeyboardAndMouse::KeyboardAndMouse(graphics::AnimatedSprite* const mouseCursor) : KeyboardAndMouse()
	{
		this->mouseCursor = mouseCursor;
	}
	KeyboardAndMouse::~KeyboardAndMouse()
	{
		if(mouseCursor)
			delete this->mouseCursor;
	}

	// Joystick
	Joystick::Joystick(LPDIRECTINPUTDEVICE8 const gc) : dev(gc)
	{ 
		// get game controller name
		DIDEVICEINSTANCE deviceInfo;
		deviceInfo.dwSize = sizeof(DIDEVICEINSTANCE);
		dev->GetDeviceInfo(&deviceInfo);
		name = deviceInfo.tszInstanceName;

		// get game controller capabilities
		DIDEVCAPS capabilities;
		capabilities.dwSize = sizeof(DIDEVCAPS);
		dev->GetCapabilities(&capabilities);

		numberOfAxes = capabilities.dwAxes;
		numberOfPOVs = capabilities.dwPOVs;
		numberOfButtons = capabilities.dwButtons;

		// clean joystick states
		ZeroMemory(&previousState, sizeof(DIJOYSTATE));
		ZeroMemory(&currentState, sizeof(DIJOYSTATE));
	}

	Joystick::~Joystick()
	{ }

	// Input Handler
	InputHandler::InputHandler(core::DirectXApp* const dxApp, const std::wstring& keyBindingsFileKeyboard, const std::wstring& keyBindingsFileJoystick, const std::wstring& keyBindingsFileGamepad) : keyBindingsFileKeyboard(keyBindingsFileKeyboard), keyBindingsFileJoystick(keyBindingsFileJoystick), keyBindingsFileGamepad(keyBindingsFileGamepad), dxApp(dxApp), listen(false)
	{
		// initialize keyboard and mouse
		kbm = new KeyboardAndMouse();
		dxApp->activeMouse = true;
		dxApp->activeKeyboard = true;

		// initialize DirectInput 8
		util::Expected<void> result = initializeDirectInput();
		if (!result.isValid())
			throw result;
		
		util::ServiceLocator::getFileLogger()->print<util::SeverityType::info>("The input handler was successfully initialized.");
	};
	InputHandler::~InputHandler()
	{
		// clear active key map
		activeKeyMap.clear();

		// clear key maps
		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
		{
			// clear keyboard key map
			for (auto x : keyMapKeyboard)
			{
				if (x.second != nullptr)
					delete x.second;
			}
			keyMapKeyboard.clear();
		}
		else if (dxApp->activeJoystick)
		{
			// clear joystick key map
			for (auto x : keyMapJoystick)
			{
				if (x.second != nullptr)
					delete x.second;
			}
			keyMapJoystick.clear();
		}
		else if (dxApp->activeGamepad)
		{
			// clear gamepad key map
			for (auto x : keyMapGamepad)
			{
				if (x.second != nullptr)
					delete x.second;
			}
			keyMapGamepad.clear();
		}
		
		// delete the pointer to the keyboard and mouse class
		delete kbm;

		// release controllers
		for (auto controller : gameControllers)
		{	
			controller->Unacquire();
			controller->Release();
		}
		gameControllers.clear();

		// release the main DirectInput device
		dev->Release();
				
		util::ServiceLocator::getFileLogger()->print<util::SeverityType::info>("The input handler was shutdown successfully.");
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////// Initialization ////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////// Polling ///////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> InputHandler::acquireInput()
	{
		if(dxApp->activeKeyboard || dxApp->activeMouse)
			// get keyboard and mouse state
			getKeyboardAndMouseState();

		if (dxApp->activeJoystick)
		{	
			// handle errors
			HRESULT hr;

			// poll the joystick
			hr = joystick->poll();

			if (hr == E_FAIL)
				return std::runtime_error("Critical error: Unable to poll the joystick device!");
		}

		if(dxApp->activeGamepad)
			// get gamepad state
		{ }

		// update the key maps
		return update();
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////// Update ////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> InputHandler::update()
	{
		// clear out any active bindings from the last frame
		bool isActive = false;
		activeKeyMap.clear();

		std::unordered_multimap<GameCommands, GameCommand*>* keyMap = nullptr;

		// get keyMap
		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyMap = &keyMapKeyboard;
		else if (dxApp->activeJoystick)
			// joystick input
			keyMap = &keyMapJoystick;
		else if (dxApp->activeGamepad)
			// gamepad input
			keyMap = &keyMapGamepad;

		if (keyMap == nullptr)
			return std::runtime_error("Critical error: Unable to find input device!");


		// loop through the map of all possible actions and find the active key bindings
		for (auto x : *keyMap)
		{
			if (x.second->chord.empty())
				continue;

			// test chord
			isActive = true;
			for (auto y : x.second->chord)
			{
				if (y.keyCode >= 0 && y.keyCode < 256)
				{
					// this is a keyboard or mouse chord
					if (getKeyState(y.keyCode) != y.keyState)
					{
						isActive = false;
						break;
					}
				}
				else if (y.keyCode >= 256 && y.keyCode < 293 && dxApp->activeJoystick)
				{
					// this is a joystick chord
					JoystickButtons jb = (JoystickButtons) (y.keyCode % 256);
					if (joystick->getButtonState(jb) != y.keyState)
					{
						isActive = false;
						break;
					}
				}
			}
			
			if (isActive)
				activeKeyMap.insert(std::pair<GameCommands, GameCommand&>(x.first, *x.second));
		}

		// if there is an active key map
		if (!activeKeyMap.empty())
		{
			// notify the currently active game states to handle the input
			if (!notify(this, false).wasSuccessful())
				return std::runtime_error("Critical error: game state was unable to handle user input!");
		}
		else
		{
			bool firstTime = true;
			if (firstTime)
			{
				// reset left mouse button
				kbm->currentState[1] = 0;
				kbm->previousState[1] = 0;
				firstTime = false;
			}

			if (listen)
			{
				// we are listening to specially requested user input
				newChordBindInfo.clear();

				// give the user the ability to "unbind" a key by pressing the "ESCAPE" key
				if (isPressed(VK_ESCAPE))
				{
					listen = false;			// stop listening ; produce normal input again
					notify(this, true);		// true: was listening to specially requested user input
					return {};				// all done
				}

				// now loop through all possible keys and check for changes
				for (unsigned int i = 0; i < 293; i++)
				{
					if (i >= 0 && i < 256)
					{
						// this is a keyboard or mouse key

						// we don't care which one of the shift or ctrl keys was pressed
						if (i >= 160 && i <= 165)
							continue;

						// push the keys the user is holding down to the chord
						if (getKeyState(i) == KeyState::StillPressed)
						{
							newChordBindInfo.push_back(BindInfo(i, getKeyState(i)));
							continue;
						}

						// now add those keys that have been pressed
						if (kbm->currentState[i] != kbm->previousState[i])	// only listen to key state changes
							//&& getKeyState(1) != KeyState::JustReleased)	// ignore when the left mouse button is released (as the menu is accessed via left mouse button click)	
							newChordBindInfo.push_back(BindInfo(i, getKeyState(i)));
					}
					else if(dxApp->activeJoystick)
					{
						// this is a joystick key
						JoystickButtons button = (JoystickButtons) (i % 256);

						if (button == JoystickButtons::EndButtons || button == JoystickButtons::EndPOV)
							continue;

						if (joystick->getButtonState(button) == KeyState::StillPressed)
						{
							newChordBindInfo.push_back(BindInfo(i, joystick->getButtonState(button)));
							continue;
						}
						
						// now add those keys that have been pressed
						if (joystick->wasPressed(button) != joystick->isPressed(button))
							newChordBindInfo.push_back(BindInfo(i, joystick->getButtonState(button)));
					}
				}

				// if there is a new chord, we have to make sure that we are not overwriting an already existing chord
				if (!newChordBindInfo.empty())
				{
					// check for new chord to not overwrite other commands
					bool newChord = true;
					for (auto x : *keyMap)
					{
						// no chord at all -> continue
						if (x.second->chord.empty())
							continue;

						// different sizes -> can't be the same chord -> continue
						if (x.second->chord.size() != newChordBindInfo.size())
							continue;
						else
						{
							// check all key bindings
							bool allTheSame = true;
							for (unsigned int i = 0; i < newChordBindInfo.size(); i++)
							{
								if (x.second->chord[i].keyCode != newChordBindInfo[i].keyCode)
								{
									// the keys are different
									allTheSame = false;
									break;
								}
								else
								{
									// the keys are the same; check for their states
									if (x.second->chord[i].keyState != newChordBindInfo[i].keyState)
									{
										// the states are different -> check for pressed <-> released mismatch
										if (x.second->chord[i].keyState == KeyState::JustPressed && newChordBindInfo[i].keyState == KeyState::JustReleased)
										{
											// do nothing
											continue;
										}
										allTheSame = false;
										break;
									}
								}
							}
							if (allTheSame)
								newChord = false;
						}
					}


					if (!newChord)
						// the just pressed chord is already bound to a command -> clear and restart
						newChordBindInfo.clear();
					else
					{
						// we have a new chord ; notify if at least one of the keys was released, else: continue
						// this is necessary to capture multipe key presses, such as "CTRL" + A
						bool sendNotification = false;
						for (auto x : newChordBindInfo)
						{
							if (x.keyCode >= 0 && x.keyCode < 256)
							{
								// this is a keyboard or mouse button
								if (getKeyState(x.keyCode) == KeyState::JustReleased)
								{
									sendNotification = true;
									listen = false;
									break;
								}
							}
							else
							{
								// this is a joystick button
								JoystickButtons button = (JoystickButtons)(x.keyCode % 256);
								if (joystick->getButtonState(button) == KeyState::JustReleased)
								{
									sendNotification = true;
									listen = false;
									break;
								}
							}
						}

						if (sendNotification)
						{
							for (auto& x : newChordBindInfo)
							{
								if (x.keyState == KeyState::JustReleased)
									x.keyState = KeyState::JustPressed;
							}
							notify(this, true);
						}
					}
				}
			}
		}

		return {};
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// Keyboard and Mouse ////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	void InputHandler::getKeyboardAndMouseState()
	{
		if (dxApp->activeKeyboard || dxApp->activeMouse)
		{
			// store the old keyboard state
			kbm->previousState = kbm->currentState;

			// read the current keyboard state
			for (int i = 0; i < 256; i++)
				kbm->currentState[i] = isPressed(i);
		}

		if (dxApp->activeMouse)
		{
			// get the position of the mouse cursor
			POINT cursorPos;
			GetCursorPos(&cursorPos);
			kbm->mouseX = cursorPos.x;
			kbm->mouseY = cursorPos.y;

			// set the cursor position
			if (kbm->mouseCursor != nullptr)
				kbm->mouseCursor->setPosition((float)kbm->mouseX, (float)kbm->mouseY);
		}
	}
	const KeyState InputHandler::getKeyState(const unsigned int keyCode) const
	{
		if (kbm->previousState[keyCode] == 1)
			if (kbm->currentState[keyCode] == 1)
				return KeyState::StillPressed;
			else
				return KeyState::JustReleased;
		else
			if (kbm->currentState[keyCode] == 1)
				return KeyState::JustPressed;
			else
				return KeyState::StillReleased;
	}

	void InputHandler::drawMouseCursor() const
	{
		if(dxApp->activeMouse)
			if(kbm->mouseCursor != nullptr)
				kbm->mouseCursor->draw();
	}

	void InputHandler::changeMouseCursorAnimationCycle(const unsigned int cycle) const
	{
		if(dxApp->activeMouse)
			if (kbm->mouseCursor != nullptr)
				this->kbm->mouseCursor->changeAnimation(cycle);
	}

	void InputHandler::updateMouseCursorAnimation(const double deltaTime) const
	{
		if(dxApp->activeMouse)
			if (kbm->mouseCursor != nullptr)
				this->kbm->mouseCursor->updateAnimation(deltaTime);
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////// DirectInput / Joystick ////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	// initialize Direct Input
	util::Expected<void> InputHandler::initializeDirectInput()
	{
		// read the configuration file
		readConfigFile();

		// initialize the main DirectInput 8 device
		if (FAILED(DirectInput8Create(dxApp->getApplicationInstance(), DIRECTINPUT_VERSION, IID_IDirectInput8, (void **)&dev, NULL)))
			return std::runtime_error("Critical error: Unable to create the main DirectInput 8 COM object!");

		// enumerate all available game controllers
		if (FAILED(dev->EnumDevices(DI8DEVCLASS_GAMECTRL, &staticEnumerateGameControllers, this, DIEDFL_ATTACHEDONLY)))
			return std::runtime_error("Critical error: Unable to enumerate input devices!");

		// if no controllers are available, there is nothing left to do, the game will have to run with keyboard and mouse only
		if (gameControllers.empty())
		{
			if (dxApp->activeJoystick && !dxApp->activeGamepad)
			{
				// load keyboard controls
				dxApp->activeJoystick = false;
				loadGameCommands();
			}
		}

		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// no joystick or gamepad support desired -> break
			return {};

		// if there are game controllers, we will have to set them up
		// this will have to be redone each time the user selects a new controller
		// for now we select the first controller as the active controller
		currentlyActiveGameController = 0;
		util::Expected<void> result = initializeGameController();
		if(!result.isValid())
			return std::runtime_error("Critical error: Unable to initialize game controller!");

		// return success
		return { };
	}

	// enumerate available game controllers
	BOOL CALLBACK InputHandler::staticEnumerateGameControllers(LPCDIDEVICEINSTANCE devInst, LPVOID pvRef)
	{
		InputHandler* inputHandlerInstance = (InputHandler*)pvRef;
		return inputHandlerInstance->enumerateGameControllers(devInst);
	}
	BOOL InputHandler::enumerateGameControllers(LPCDIDEVICEINSTANCE devInst)
	{
		// enumerate devices
		LPDIRECTINPUTDEVICE8 gameController;

		// create interface for the current game controller
		if (FAILED(dev->CreateDevice(devInst->guidInstance, &gameController, NULL)))
			return DIENUM_CONTINUE;
		else
		{
			// store the game controller
			gameControllers.push_back(gameController); // (std::make_pair<std::wstring, LPDIRECTINPUTDEVICE8>(devInst->tszProductName, std::move(gameController)));
			return DIENUM_CONTINUE;
		}
	}

	// initialize game controller
	util::Expected<void> InputHandler::initializeGameController()
	{
		if (currentlyActiveGameController < 0 || currentlyActiveGameController >= gameControllers.size())
			return std::runtime_error("Critical error: Game controller index out of range!");

		// get currently active game controller
		LPDIRECTINPUTDEVICE8 gameController = gameControllers[currentlyActiveGameController];

		// set cooperative level
		//  - DISCL_BACKGROUND: receive notifications when the application is in the background as well as in the foreground
		//  - DISCL_EXCLUSIVE: exclusive access: no other application can request exclusive access to the game controller
		if (FAILED(gameController->SetCooperativeLevel(dxApp->getMainWindow(), DISCL_BACKGROUND | DISCL_EXCLUSIVE)))
			return std::runtime_error("Critical error: Unable to set the cooperative level for the game controller!");

		// set data format
		if (FAILED(gameController->SetDataFormat(&c_dfDIJoystick)))
			return std::runtime_error("Critical error: Unable to set data format for the game controller!");

		// set range and dead zone of joystick axes
		if (FAILED(gameController->EnumObjects(&staticSetGameControllerAxesRanges, gameController, DIDFT_AXIS)))
			throw "Critical error: Unable to set axis ranges of game controllers!";

		// acquire the controller
		if (FAILED(gameController->Acquire()))
			throw "Critical error: Unable to acquire the game controller!";

		// create the joystick class object
		if (joystick != nullptr)
			delete joystick;
		joystick = new Joystick(gameController);

		// return success
		return { };
	}

	// set controller axes
	BOOL CALLBACK InputHandler::staticSetGameControllerAxesRanges(LPCDIDEVICEOBJECTINSTANCE devObjInst, LPVOID pvRef)
	{
		// the game controller
		LPDIRECTINPUTDEVICE8 gameController = (LPDIRECTINPUTDEVICE8)pvRef;
		gameController->Unacquire();
		
		// structure to hold game controller range properties
		DIPROPRANGE gameControllerRange;

		// set the range to -100 and 100
		gameControllerRange.lMin = -100;
		gameControllerRange.lMax = 100;

		// set the size of the structure
		gameControllerRange.diph.dwSize = sizeof(DIPROPRANGE);
		gameControllerRange.diph.dwHeaderSize = sizeof(DIPROPHEADER);

		// set the object that we want to change		
		gameControllerRange.diph.dwHow = DIPH_BYID;
		gameControllerRange.diph.dwObj = devObjInst->dwType;

		// now set the range for the axis		
		if (FAILED(gameController->SetProperty(DIPROP_RANGE, &gameControllerRange.diph))) {
			return DIENUM_STOP;
		}

		// structure to hold game controller axis dead zone
		DIPROPDWORD gameControllerDeadZone;

		// set the dead zone to 1%
		gameControllerDeadZone.dwData = 100;

		// set the size of the structure
		gameControllerDeadZone.diph.dwSize = sizeof(DIPROPDWORD);
		gameControllerDeadZone.diph.dwHeaderSize = sizeof(DIPROPHEADER);

		// set the object that we want to change
		gameControllerDeadZone.diph.dwHow = DIPH_BYID;
		gameControllerDeadZone.diph.dwObj = devObjInst->dwType;

		// now set the dead zone for the axis
		if (FAILED(gameController->SetProperty(DIPROP_DEADZONE, &gameControllerDeadZone.diph)))
			return DIENUM_STOP;

		return DIENUM_CONTINUE;
	}

	// poll
	HRESULT Joystick::poll()
	{
		HRESULT hr;

		// store the current state
		CopyMemory(&previousState, &currentState, sizeof(DIJOYSTATE));
		ZeroMemory(&currentState, sizeof(DIJOYSTATE));

		// poll the device to read the current state
		hr = dev->Poll();

		if (FAILED(hr))
		{
			// DirectInput lost the device, try to re-acquire it
			hr = dev->Acquire();
			while (hr == DIERR_INPUTLOST)
				hr = dev->Acquire();

			// return if a fatal error is encountered
			if ((hr == DIERR_INVALIDPARAM) || (hr == DIERR_NOTINITIALIZED))
				return E_FAIL;

			// if another application has control of this device, we have to wait for our turn
			if (hr == DIERR_OTHERAPPHASPRIO)
				return S_OK;
		}

		// now if everything is okay, we can get the state of the device
		if (FAILED(hr = dev->GetDeviceState(sizeof(DIJOYSTATE), &currentState)))
			return hr;

		return S_OK;
	}
	const bool Joystick::isPressed(const JoystickButtons button) const
	{
		if (button < JoystickButtons::EndPOV)
		{
			// POV button
			switch (button)
			{
				//   0 degrees: north
				//  90 degrees: east
				// 180 degrees: south
				// 270 degrees: west
			case JoystickButtons::JoyPOV_Up:
				return currentState.rgdwPOV[0] == 0;
			case JoystickButtons::JoyPOV_Right:
				return currentState.rgdwPOV[0] == 9000;
			case JoystickButtons::JoyPOV_Down:
				return currentState.rgdwPOV[0] == 18000;
			case JoystickButtons::JoyPOV_Left:
				return currentState.rgdwPOV[0] == 27000;
			}
		}

		if (button > JoystickButtons::EndPOV && button < JoystickButtons::EndButtons)
		{	
			// regular button
			JoystickButtons btn = (JoystickButtons)(button - 5);
			return (currentState.rgbButtons[btn] & 0x80) ? 1 : 0;
		}

		return false;
	}
	const bool Joystick::wasPressed(const JoystickButtons button) const
	{
		if (button < JoystickButtons::EndPOV)
		{
			// POV button
			switch (button)
			{
				//   0 degrees: north
				//  90 degrees: east
				// 180 degrees: south
				// 270 degrees: west
			case JoystickButtons::JoyPOV_Up:
				return previousState.rgdwPOV[0] == 0;
			case JoystickButtons::JoyPOV_Right:
				return previousState.rgdwPOV[0] == 9000;
			case JoystickButtons::JoyPOV_Down:
				return previousState.rgdwPOV[0] == 18000;
			case JoystickButtons::JoyPOV_Left:
				return previousState.rgdwPOV[0] == 27000;
			}
		}

		if (button > JoystickButtons::EndPOV && button < JoystickButtons::EndButtons)
		{
			// regular button
			JoystickButtons btn = (JoystickButtons)(button - 5);
			return (previousState.rgbButtons[btn] & 0x80) ? 1 : 0;
		}
		
		return false;
	}
	const KeyState Joystick::getButtonState(const JoystickButtons button) const
	{
		if (button == JoystickButtons::EndButtons || button == JoystickButtons::EndPOV)
			return KeyState::StillReleased;

		if (wasPressed(button) == 1)
			if (isPressed(button) == 1)
				return KeyState::StillPressed;
			else
				return KeyState::JustReleased;
		else
			if (isPressed(button) == 1)
				return KeyState::JustPressed;
			else
				return KeyState::StillReleased;
	}


	// read configuration file
	void InputHandler::readConfigFile()
	{
		if (dxApp->hasValidConfigurationFile())
		{
			// configuration file exists, try to read from it
			std::wstring pathToPrefFile = dxApp->getPathToConfigurationFiles() + L"\\" +  dxApp->getPrefsFile();

			try
			{
				sol::state lua;
				lua.script_file(util::StringConverter::ws2s(pathToPrefFile));

				// read from the configuration file, default to false
				dxApp->activeJoystick = lua["config"]["joystick"].get_or(false);
				dxApp->activeGamepad = lua["config"]["gamepad"].get_or(false);
#ifndef NDEBUG
				std::stringstream res;
				res << "The game controller states were read from the Lua configuration file: joystick: " << std::boolalpha << dxApp->activeJoystick << " --- gamepad: " << std::boolalpha << dxApp->activeGamepad << ".";
				util::ServiceLocator::getFileLogger()->print<util::SeverityType::info>(res.str());
#endif
			}
			catch (std::exception)
			{
				util::ServiceLocator::getFileLogger()->print<util::SeverityType::warning>("Unable to read configuration file. Joystick and gamepad support disabled!");
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// Bindings /////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	util::Expected<void> InputHandler::saveGameCommands() const
	{
		std::ofstream keyBindingsOut;

		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyBindingsOut = std::ofstream(this->keyBindingsFileKeyboard, std::ios::out);
		else if(dxApp->activeJoystick)
			// joystick input
			keyBindingsOut = std::ofstream(this->keyBindingsFileJoystick, std::ios::out);
		else if(dxApp->activeGamepad)
			// gamepad input
			keyBindingsOut = std::ofstream(this->keyBindingsFileGamepad, std::ios::out);
		else
			return std::runtime_error("Critical error: Unable to deduce input device!");

		// open text archive
		boost::archive::text_oarchive oa(keyBindingsOut);
		
		const std::unordered_multimap<GameCommands, GameCommand*>* keyMap = nullptr;

		// get keyMap
		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyMap = &keyMapKeyboard;
		else if (dxApp->activeJoystick)
			// joystick input
			keyMap = &keyMapJoystick;
		else if (dxApp->activeGamepad)
			// gamepad input
			keyMap = &keyMapGamepad;

		if (keyMap == nullptr)
			return std::runtime_error("Critical error: Unable to deduce input device!");

		for (auto gameCommand : *keyMap)
		{
			oa << gameCommand.first;
			oa << gameCommand.second;
		}

		keyBindingsOut.close();

		// return success
		return { };
	}

	util::Expected<void> InputHandler::loadGameCommands()
	{
		// create default key bindings if file does not yet exist
		if (!dxApp->activeGamepad && !dxApp->activeJoystick)
		{
			if (GetFileAttributes(keyBindingsFileKeyboard.c_str()) == INVALID_FILE_ATTRIBUTES)
			{
				setDefaultKeyMap();
				saveGameCommands();
			}
		}
		else if (dxApp->activeJoystick)
		{
			if (GetFileAttributes(keyBindingsFileJoystick.c_str()) == INVALID_FILE_ATTRIBUTES)
			{
				setDefaultKeyMap();
				saveGameCommands();
			}
		}
		else if (dxApp->activeGamepad)
		{
			if (GetFileAttributes(keyBindingsFileGamepad.c_str()) == INVALID_FILE_ATTRIBUTES)
			{
				setDefaultKeyMap();
				saveGameCommands();
			}
		}

		std::unordered_multimap<GameCommands, GameCommand*>* keyMap = nullptr;
		// get keyMap
		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyMap = &keyMapKeyboard;
		else if (dxApp->activeJoystick)
			// joystick input
			keyMap = &keyMapJoystick;
		else if (dxApp->activeGamepad)
			// gamepad input
			keyMap = &keyMapGamepad;

		if (keyMap == nullptr)
			return std::runtime_error("Critical error: Unable to find input device!");

		// clear map
		keyMap->clear();

		// load data
		GameCommands gcs;

		std::ifstream keyBindingsIn;

		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyBindingsIn = std::ifstream(this->keyBindingsFileKeyboard, std::ios::in);
		else if (dxApp->activeJoystick)
			// joystick input
			keyBindingsIn = std::ifstream(this->keyBindingsFileJoystick, std::ios::in);
		else if (dxApp->activeGamepad)
			// gamepad input
			keyBindingsIn = std::ifstream(this->keyBindingsFileGamepad, std::ios::in);
		else
			return std::runtime_error("Critical error: Unable to deduce input device!");
		
		// open the file
		boost::archive::text_iarchive ia(keyBindingsIn);
		
		// populate map with information from file
		while (!keyBindingsIn.eof())
		{
			ia >> gcs;
			ia >> unmarshalledGameCommand;
			keyMap->insert(std::pair<GameCommands, GameCommand*>(gcs, unmarshalledGameCommand));
		}

		// close the file
		keyBindingsIn.close();

		// return success
		return { };
	}

	void InputHandler::resetKeyStates()
	{
		this->kbm->currentState.fill(0);
		this->kbm->previousState.fill(0);
	}

	void InputHandler::enableListening()
	{
		this->resetKeyStates();
		listen = true;
	}

	void InputHandler::disableListening()
	{
		listen = false;
		this->resetKeyStates();
	}

	util::Expected<void> InputHandler::getKeysMappedToCommand(const GameCommands gameCommand, std::vector<std::vector<BindInfo> >& map) const
	{
		const std::unordered_multimap<GameCommands, GameCommand*>* keyMap = nullptr;

		// get keyMap
		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyMap = &keyMapKeyboard;
		else if (dxApp->activeJoystick)
			// joystick input
			keyMap = &keyMapJoystick;
		else if (dxApp->activeGamepad)
			// gamepad input
			keyMap = &keyMapGamepad;

		if (keyMap == nullptr)
			return std::runtime_error("Critical error: Unable to find input device!");

		std::vector<std::vector<BindInfo> > vecBI;
		for(auto x : *keyMap)
			if (x.first == gameCommand)
				map.push_back(x.second->chord);

		// return success
		return { };
	}

	util::Expected<void> InputHandler::getCommandsMappedToGameAction(const GameCommands gameCommand, std::vector<GameCommand*>& commands) const
	{
		const std::unordered_multimap<GameCommands, GameCommand*>* keyMap = nullptr;

		// get keyMap
		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyMap = &keyMapKeyboard;
		else if (dxApp->activeJoystick)
			// joystick input
			keyMap = &keyMapJoystick;
		else if (dxApp->activeGamepad)
			// gamepad input
			keyMap = &keyMapGamepad;

		if (keyMap == nullptr)
			return std::runtime_error("Critical error: Unable to find input device!");

		for (auto x : *keyMap)
			if (x.first == gameCommand)
				commands.push_back(x.second);

		// return success
		return { };
	}

	util::Expected<void> InputHandler::insertNewCommand(GameCommands gameCommand, GameCommand& command)
	{
		std::unordered_multimap<GameCommands, GameCommand*>* keyMap = nullptr;

		// get keyMap
		if (!dxApp->activeJoystick && !dxApp->activeGamepad)
			// keyboard input
			keyMap = &keyMapKeyboard;
		else if (dxApp->activeJoystick)
			// joystick input
			keyMap = &keyMapJoystick;
		else if (dxApp->activeGamepad)
			// gamepad input
			keyMap = &keyMapGamepad;

		if (keyMap == nullptr)
			return std::runtime_error("Critical error: Unable to find input device!");

		keyMap->insert(std::pair<GameCommands, GameCommand*>(gameCommand, &command));

		// return succcess
		return { };
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////// Key Codes ////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	const util::Expected<std::wstring> InputHandler::getKeyName(const unsigned int keyCode) const
	{
		if (keyCode >= 0 && keyCode < 256)
		{
			// this is a keyboard or mouse code

			int scanCode = MapVirtualKey(keyCode, MAPVK_VK_TO_VSC);

			// the MapVirtualKey method strips the extended bit for some keys
			// thus we have to add them again
			switch (keyCode)
			{
			case VK_LEFT: case VK_RIGHT: case VK_UP: case VK_DOWN:
			case VK_HOME: case VK_END: case VK_PRIOR: case VK_NEXT:
			case VK_INSERT: case VK_DELETE: case VK_DIVIDE: case VK_NUMLOCK:
				scanCode |= 0x100;
				break;
			}

			// now get the key name
			wchar_t keyName[256];
			if (GetKeyNameText(scanCode << 16, keyName, 256) != 0)
				return (std::wstring)keyName;
			else
			{
				HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
				if (hr == S_OK)
				{
					// might be a mouse button
					// I have no idea why they aren't caught by the above code
					switch (keyCode)
					{
					case 1:
						return (std::wstring)L"Left Mouse Button";
					case 2:
						return (std::wstring)L"Right Mouse Button";
					case 4:
						return (std::wstring)L"Middle Mouse Button";
					case 5:
						return (std::wstring)L"Extra Mouse Button 1";
					case 6:
						return (std::wstring)L"Extra Mouse Button 2";

					default:
						return (std::wstring)L"Unknown Key";
					}

				}
				else
					return hr;
			}
		}
		
		if (keyCode >= 256 && keyCode < 293)
		{
			// joystick button
			unsigned int button = keyCode % 256;
			return (std::wstring)enumToString(JoystickButtons(button));
		}

		return (std::wstring)L"Unknwon Key";
		
	}
}